-- AddEmote
-- by Travis S. Casey (efindel@gmail.com)

-- create a namespace for our functions and variables
AddEmote = {}
-- initialize a few variables
AddEmote.offset = 0
AddEmote.numEmotes = 0
AddEmote.Debug = false

AddEmote.numButtons = 7

-- create table to hold the emotes, with a few example ones
AddEmote_Emotes = {}

-- the structure of an Emote entry
-- AddEmote_Emotes[emotename] = {
--   ["type"] = <one of "EMOTE", "SAY", "YELL"
--   ["text"] = "the text for the emote"
-- }
--
-- If an emote is character-specific, we mangle the name:
-- emotename_charactername-realm, or
-- emotename_charactername

-- debugging functions
local function DebugPrint(string)
  if AddEmote.Debug then
    print(string)
  end
end

-- doesn't seem to be working right. REMOVE?
local function trim(string)
  return (string.gsub("^%s*(.-)%s*$", "%1"))
end

-- utility function for iterating through a table in key order
-- some code borrowed from wowwiki
-- utility function to sort a table by the keys, returns an iterator
-- usage:  for title, value in AddEmote.PairsByKeys(sometable) do ... end
function AddEmote.PairsByKeys (t, f)
  local a = {}
  local n
  
  for n in pairs(t) do table.insert(a, n) end
  table.sort(a,f)
  local i = 0
  local iter = function()
    i = i+1
    if a[i] == nil then
      return nil
    else
      return a[i], t[a[i]]
    end
  end
  return iter
end

local PairsByKeys = AddEmote.PairsByKeys

local thistoonshort = "_"..GetUnitName("player")
local thistoon = thistoonshort.."-"..GetRealmName()

-- utility function for character-specific emotes
function AddEmote.GetShortName (emotename)
  return string.gsub(emotename, "_.*", "", 1)
end

-- get the data for an emote, letting character-specific ones
-- take priority
function AddEmote.GetEmote (emotename)
  if AddEmote_Emotes[emotename..thistoon] ~= nil then
    return AddEmote_Emotes[emotename..thistoon]
  elseif AddEmote_Emotes[emotename..thistoonshort] ~= nil then
    return AddEmote_Emotes[emotename..thistoonshort]
  else
    return AddEmote_Emotes[emotename]
  end
end

function AddEmote.EmoteText (emotename)
  local which = AddEmote.GetEmote(emotename)
  if which then
    return which["text"]
  end
end

-- get a range of emotes, sorted as a list, from a to b. 
-- "output" is a table to hold the list
function AddEmote.GetEmotesAsTable (output, a, b)
  local entry
  local t = {}
  local n = 1

  wipe(output)
  for entry in pairs(AddEmote_Emotes) do table.insert(t, entry) end
  table.sort(t)
  for _, entry in pairs(t) do 
    if ((n >= a) and (n <= b)) then
      table.insert(output, entry)
    end
    n = n + 1
  end
  return output
end

function AddEmote.ScrollUpdate()
  local shownemotes = {}
  local lastindex = nil

  DebugPrint("ScrollUpdate() called")
  DebugPrint("offset: "..AddEmote.offset)
  -- get the emotes we're going to show
  shownemotes = AddEmote.GetEmotesAsTable(shownemotes, AddEmote.offset, (AddEmote.offset+AddEmote.numButtons))
  -- set the number we have to scroll through
  AddEmote.scrollbar:SetMinMaxValues(0, AddEmote.numEmotes)
  -- iterate through the list, showing the ones in the range we've been given (AddEmote.offset)
  for index = 1, AddEmote.numButtons do
    local emoteframe = _G["AddEmoteEmote"..index]
    if shownemotes[index] ~= nil then
      DebugPrint("showing emote: "..shownemotes[index])
      local emoteinfo = AddEmote_Emotes[shownemotes[index]]
      lastindex = index
      emoteframe.Slash:SetText(shownemotes[index])
      emoteframe.Slash:SetCursorPosition(0) -- workaround for Blizzard bug where text is set, but doesn't show
      emoteframe.Type.Text:SetText(emoteinfo.type)
      emoteframe.Text:SetText(emoteinfo.text)
      emoteframe.Text:SetCursorPosition(0)
      emoteframe:Show()
    else
      -- if we don't have an emote for this position, hide that frame
      emoteframe:Hide()
    end
  end
  -- ... but if we have blank frames, make the first one show up blank, so the
  -- user can use it to add new emotes
  if AddEmote.numEmotes == 0 then lastindex = 0 end
  if lastindex ~= nil and lastindex < AddEmote.numButtons then
    lastindex = lastindex + 1
    emoteframe = _G["AddEmoteEmote"..lastindex]
    emoteframe.Slash:SetText("")
    emoteframe.Type.Text:SetText("EMOTE")
    emoteframe.Text:SetText("")
    emoteframe:Show()
  end
end

function AddEmote.OnMouseWheel(self, delta)
  local current = AddEmote.offset

  if (delta < 0) and (current < AddEmote.numEmotes) then
    AddEmote.scrollbar:SetValue(current + 1)
  elseif (delta > 0) and (current > 0) then
    AddEmote.scrollbar:SetValue(current - 1)
  end
end

function AddEmote.DropDownOnClick(self, arg1, arg2)
  arg2.Text:SetText(arg1)
end

function AddEmote.DropDownInitialize(frame)
  local info = UIDropDownMenu_CreateInfo()

  info.func = AddEmote.DropDownOnClick
  info.arg2 = frame
  info.notCheckable = true
  info.justifyH = "CENTER"

  info.text = "EMOTE"; info.arg1 = info.text
  UIDropDownMenu_AddButton(info)
  info.text = "SAY"; info.arg1 = info.text
  UIDropDownMenu_AddButton(info)
  info.text = "YELL"; info.arg1 = info.text
  UIDropDownMenu_AddButton(info)
end

-- okay, here's a big one.  When we have a bunch of alternatives,
-- figure out which one we want to use.  The logic is:
--   if there aren't any alternatives, return the one thing we have
--   check the alternatives to see if we have all the units needed
--   if we do, add it to the list of ones we found valid, with a
--     rank according to how many units it used
--   if none of them were valid, return everything, and let the
--     "everyone", "no class", etc. fallbacks handle it
--   if some were valid, take the highest ranked alternatives and
--     return them for the caller to pick from
function AddEmote.GetValidStrings (emotename)
  local text = AddEmote.EmoteText(emotename)

  -- if there's only one option, get back out
  -- but return it as a table, so the "unpack" will succeed
  if not string.find(text, "|") then
    return { text }
  end

  local emotestrings = { strsplit("|", text) }
  local tempstrings = {}
  local i, line, count, maxcount, foundvalid
  maxcount = 0
  foundvalid = false

  for i, line in ipairs(emotestrings) do
    if line ~= "" then 
--      line = trim(line) -- remove leading and trailing whitespace
      DebugPrint("AddEmote considering: "..line)
      count = ChatSubs.ValidUnits(line)
      if count then
        foundvalid = true
        DebugPrint("valid - "..count.." units")
        if count > maxcount then
          maxcount = count
        end
        if not tempstrings[count] then 
          tempstrings[count] = {} 
        end
        table.insert(tempstrings[count], line)
      end
    end
  end

  -- if tempstrings is empty, nothing was valid... so we just try anything
  if not foundvalid then
    return emotestrings
  else
    -- okay, we have valid tempstrings.  Now, we want to use the
    -- set of valid alternatives with the most variables
    return tempstrings[maxcount]
  end
end

-- we just created a new emote, or loaded an existing one at load time
-- now, make it be a /command
function AddEmote.Register (emotename, nomessage)
  if ( nomessage ~= true ) then
    print("AddEmote: registered emote "..emotename)
  end
  emotename = AddEmote.GetShortName(emotename)
  SlashCmdList["ADDEMOTE_"..emotename] = function(text)
     ChatSubs.SetText(text)
    SendChatMessage( GetRandomArgument(unpack(AddEmote.GetValidStrings(emotename))), AddEmote.GetEmote(emotename).type )
    ChatSubs.SetText("")
  end
  _G["SLASH_ADDEMOTE_"..emotename.."1"] = "/"..emotename
  AddEmote.numEmotes = AddEmote.numEmotes + 1
end

function AddEmote.UnRegister (emotename, nomessage)
  -- you should delete the AddEmote_Emote entry before calling this. 
  -- if there is still an entry, don't do anything.
  if AddEmote.GetEmote(emotename) ~= nil then
    print("AddEmote: other emote still exists for /"..emotename..", not removing slash command.")
    return
  else
    print("AddEmote: removing slash command for /"..emotename..".")
    SlashCmdList["ADDEMOTE_"..emotename] = nil
    _G["SLASH_ADDEMOTE_"..emotename.."1"] = nil
  end
end

-- add an emote
function AddEmote.AddOne (emotename, emotestring, type)
  DebugPrint(emotename..", "..emotestring..", "..type)
  AddEmote_Emotes[emotename] = {}
  AddEmote_Emotes[emotename].text = emotestring
  if type then 
    AddEmote_Emotes[emotename].type = type
  end
  AddEmote.Register(emotename)
end

-- set up the /addemote command
SlashCmdList["ADDEMOTE_ADDEMOTE"] = function(msg)
  AddEmote.ScrollUpdate()
  InterfaceOptionsFrame_OpenToCategory("AddEmote")
  InterfaceOptionsFrame_OpenToCategory("AddEmote")
end

-- delete an emote
function AddEmote.DeleteEmote (emotename)
  AddEmote_Emotes[emotename] = nil
  -- let's see whether there's still an emote with
  -- this command name.
  local slash = AddEmote.GetShortName(emotename)
  AddEmote.UnRegister(slash) -- unregister won't actually unregister if the command still has uses
  AddEmote.numEmotes = AddEmote.numEmotes - 1
end

-- actually add the command to the game
SLASH_ADDEMOTE_ADDEMOTE1 = "/addemote"

-- at login time, we need to turn all the saved emotes into
-- slash commands.  This does that.
function AddEmote.OnEvent()
  -- the only event we respond to is VARIABLES_LOADED, 
  -- so we don't have to check which event this is
  local name, value

  for name, value in pairs(AddEmote_Emotes) do
    if type(value) == "string" then
      -- this is data from a previous version, convert to 3.x
      AddEmote_Emotes[name] = {}
      AddEmote_Emotes[name].type = "EMOTE"
      AddEmote_Emotes[name].text = value
    end
    AddEmote.Register(name, true)
  end

  if AddEmote_CharacterEmotes then
    -- we have character specific emotes from a previous version.
    -- convert them!
    for name, value in pairs(AddEmote_CharacterEmotes) do
      name = name..thistoon
      if type(value) == "string" then
        -- this is data from a previous version, convert to 3.x
        AddEmote_Emotes[name] = {}
        AddEmote_Emotes[name].type = "EMOTE"
        AddEmote_Emotes[name].text = value
      end
      AddEmote.Register(name, true)
    end
  end

  if AddEmote.numEmotes == 0 then
    -- there were no emotes to load, so this must be a fresh install.
    -- we'll add a few sample emotes.
    AddEmote.AddOne("challenge", "I challenge you, <target>!||I'm looking for a challenge!", "SAY")
    AddEmote.AddOne("chinrub", "rubs <hisher> chin thoughtfully.", "EMOTE")
    AddEmote.AddOne("munch", "gets out <text> and munches on it.||gets something out and munches on it.", "EMOTE")
    AddEmote.AddOne("pet", "pets <target> <text>.||looks around for something to pet.", "EMOTE")
  end

  -- set up some things we couldn't at load time, since the frames
  -- might not have existed yet
  AddEmote.scrollbar = AddEmoteEmotesListScrollbar
  AddEmote.scrollbar:SetMinMaxValues(0, AddEmote.numEmotes)
  AddEmote.scrollbar:SetValueStep(1)
  AddEmote.scrollbar:SetValue(0)
  AddEmote.scrollbar:SetStepsPerPage(AddEmote.numButtons)

  AddEmoteUsage:SetText([[You can use substitution strings in emotes of the form |cFFFFFF00<whowhat>|r. |CFFFFFF00who|r: |CFF00FF00player|r, target, pet. (Default is in |cFF00FF00green|r)
|cFFFFFF00what|r: |CFF00FF00name|r, race, class, type, sex, heshe, himher, hisher, hishers, boygirl, manwoman, lordlady, sirmadam.
|CFFFFFF00Example|r: 'scratches <hisher> chin.' would show 'his' or 'her' depending on the sex of the player character.
You can list multiple alternatives, separating them with the '||' symbol.
|CFFFFFF00Example|r: 'scratches <target> on <targethisher> back.||scratches <hisher> own back.'
would show the first if you have a target, the latter without one.
|cFFFFFF00<text>|r is a placeholder for text you input with the /command.
|CFFFFFF00Example|r: create an emote named '/add' with '<target>, <text>' to address someone you're targetting by name.
To create character-specific emotes, add "_Name" or "_Name-Realmname" to the command name.
See the README in Interface/Addons/AddEmote for more information.]])

  AddEmote.ScrollUpdate()

  print("AddEmote: variables loaded.")
end

function AddEmote.OnLoad()
  -- let's set up the interface!
  AddEmote.panel = _G["AddEmoteUI"]
  -- register it for the interface options GUI
  AddEmote.panel.name = "AddEmote"
  InterfaceOptions_AddCategory(AddEmoteUI)
  AddEmote.Version = GetAddOnMetadata("AddEmote", "Version")
  _G["AddEmoteUIVersion"]:SetText(AddEmote.Version)

  AddEmote.panel:RegisterEvent("PLAYER_LOGIN")
  AddEmote.panel:SetScript("OnEvent", AddEmote.OnEvent)
  
  AddEmote.scrollbar = _G["AddEmoteEmotesListScrollbar"]
  -- we actually finish setting up the scrollbar after loading emotes
end
